<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2018 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2018.2/1705499
 */

namespace Api\Controller;

use Api\AbstractApiController;
use P4\Spec\Change;
use P4\Spec\Exception\NotFoundException;
use Projects\Model\Project;
use Reviews\UpdateService;
use Workflow\Manager;
use Zend\View\Model\JsonModel;
use Api\Converter\Reviewers as ReviewersConverter;
use Zend\Http\Response;
use Api\Filter\Changes;

/**
 * API controller providing a service for changes
 *
 * @SWG\Resource(
 *   apiVersion="v9",
 *   basePath="/api/v9/"
 * )
 */
class ChangesController extends AbstractApiController
{

    /**
     * @SWG\Api(
     *     path="changes/{change}/defaultreviewers",
     *     @SWG\Operation(
     *         method="GET",
     *         summary="Get default reviewers for a given change id.",
     *         notes="All authenticated users are able to use this API."
     *     )
     * )
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     * {
     *      "change": {
     *          "id": "1050",
     *          "defaultReviewers": {
     *              "groups": {
     *                  "group1": {"required": "1"},
     *                  "group2": {}
     *              },
     *              "users": {
     *                  "user1": {},
     *                  "user2": {"required": "true"}
     *              }
     *          }
     *      }
     * }
     *
     * @return JsonModel
     * @throws \P4\Exception
     */
    public function defaultReviewersAction()
    {
        $error    = null;
        $services = $this->getServiceLocator();
        $services->get('permissions')->enforce('authenticated');
        $changeId = $this->getEvent()->getRouteMatch()->getParam('change');
        $p4User   = $services->get('p4_user');
        // Check if the change exists.
        try {
            $change = Change::fetchById($changeId, $p4User);
        } catch (NotFoundException $e) {
            $error = $e;
        } catch (\InvalidArgumentException $e) {
            $error = $e;
        }
        if ($error) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_404);
            return new JsonModel(
                array(
                    'isValid'  => false,
                    'messages' => array($error->getMessage()),
                )
            );
        }
        $reviewers = array(); // Only variables can be passed by reference
        $reviewers = ReviewersConverter::expandUsersAndGroups(
            UpdateService::mergeDefaultReviewersForChange($change, $reviewers, $p4User)
        );

        return new JsonModel(
            array(
                'change' => array('id' => $changeId, 'defaultReviewers' => $reviewers)
            )
        );
    }

    /**
     * @SWG\Api(
     *     path="changes/{change}/affectsprojects",
     *     @SWG\Operation(
     *         method="GET",
     *         summary="Get projects, and branches, affected by a given change id.",
     *         notes="All authenticated users are able to use this API."
     *     )
     * )
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     *     {
     *         "change": {
     *             "id":"1050",
     *             "projects": {
     *                 "jam": [
     *                     "live",
     *                     "main"
     *                 ]
     *             }
     *         }
     *     }
     *
     * @return JsonModel
     * @throws \P4\Exception
     */
    public function affectsProjectsAction()
    {
        $error    = null;
        $services = $this->getServiceLocator();
        $services->get('permissions')->enforce('authenticated');
        $changeId = $this->getEvent()->getRouteMatch()->getParam('change');
        $p4       = $services->get('p4');
        // Check if the change exists.
        try {
            $change = Change::fetchById($changeId, $p4);
        } catch (NotFoundException $e) {
            $error = $e;
        } catch (\InvalidArgumentException $e) {
            $error = $e;
        }
        if ($error) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_404);
            return new JsonModel(
                array(
                    'isValid'  => false,
                    'messages' => array($error->getMessage()),
                )
            );
        }

        return new JsonModel(
            array(
                'change' => array('id' => $changeId, 'projects' => Project::getAffectedByChange($change, $p4))
            )
        );
    }

    /**
     * Carries out enforced checks on the change if enabled
     * @return JsonModel
     * @throws \Application\Config\ConfigException
     * @throws \P4\Exception
     * @throws \Record\Exception\NotFoundException
     */
    private function checkEnforced()
    {
        $changeId = $this->getEvent()->getRouteMatch()->getParam('change');
        $manager  = new Manager($this->getServiceLocator());
        return $this->translateWorkflowResult($manager->checkEnforced($changeId));
    }

    /**
     * Carries out checks on the change if enabled
     * @SWG\Api(
     *     path="changes/{id}/check",
     *     @SWG\Operation(
     *         method="GET",
     *         summary="Performs checks on the change if enabled",
     *         notes="Performs checks on the change if workflow configuration requires it",
     *         nickname="check",
     *         @SWG\Parameter(
     *             name="id",
     *             description="Change to check",
     *             paramType="form",
     *             type="string",
     *             required=true
     *         ),
     *         @SWG\Parameter(
     *             name="type",
     *             description="The type of check. Must have a value of enforced, strict or shelve",
     *             paramType="form",
     *             type="string",
     *             required=true
     *         )
     *     )
     * )
     *
     * @apiUsageExample Carry out enforced checks on a change if configured
     *
     *   To check change 42:
     *
     *   ```bash
     *   curl -u "username:password" \
     *        -X POST \
     *        "https://my-swarm-host/api/v9/changes/42/check?type=enforced"
     *   ```
     *
     *   JSON Response:
     *
     *   ```json
     *   {
     *     "status":   "OK",
     *     "isValid":  "true",
     *     "messages": []
     *   }
     *   ```
     *
     * @apiUsageExample Carry out enforced checks on a change if configured. Example of a failed 'requires review'
     *
     *   To check change 42:
     *
     *   ```bash
     *   curl -u "username:password" \
     *        -X POST \
     *        "https://my-swarm-host/api/v9/changes/42/check?type=enforced"
     *   ```
     *
     *   JSON Response:
     *
     *   ```json
     *   {
     *     "status":   "NO_REVIEW",
     *     "isValid":  "false",
     *     "messages": ["Change [42] must be associated with a review"]
     *   }
     *   ```
     * @apiUsageExample Carry out strict checks on a change if configured
     *
     *   To check change 42:
     *
     *   ```bash
     *   curl -u "username:password" \
     *        -X POST \
     *        "https://my-swarm-host/api/v9/changes/42/check?type=strict"
     *   ```
     *
     *   JSON Response:
     *
     *   ```json
     *   {
     *     "status":   "OK",
     *     "isValid":  "true",
     *     "messages": []
     *   }
     *   ```
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     *   {
     *     "status":   "OK",
     *     "isValid":  "true",
     *     "messages": []
     *   }
     *
     * @return JsonModel
     * @throws \Application\Config\ConfigException
     * @throws \P4\Exception
     * @throws \Record\Exception\NotFoundException
     */
    public function checkAction()
    {
        $result = null;
        $filter = new Changes();
        $filter->setData($this->getRequest()->getQuery()->toArray());
        $valid = $filter->isValid();
        if ($valid) {
            $type = $this->getRequest()->getQuery('type');
            switch ($type) {
                case Changes::ENFORCED:
                    $result = $this->checkEnforced();
                    break;
                case Changes::STRICT:
                    $result = $this->checkStrict();
                    break;
                case Changes::SHELVE:
                    $result = $this->checkShelve();
                    break;
            }
        } else {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_400);
            $result = $this->prepareErrorModel(
                new JsonModel(
                    array(
                        'isValid'  => false,
                        'messages' => $filter->getMessages()
                    )
                )
            );
        }
        return $result;
    }

    /**
     * Carries out shelve checks on the change if enabled (shelve-submit)
     * @return JsonModel
     * @throws \Application\Config\ConfigException
     * @throws \P4\Exception
     * @throws \Record\Exception\NotFoundException
     */
    private function checkShelve()
    {
        $changeId = $this->getEvent()->getRouteMatch()->getParam('change');
        $manager  = new Manager($this->getServiceLocator());
        return $this->translateWorkflowResult($manager->checkShelve($changeId));
    }

    /**
     * Carries out strict checks on the change if enabled.
     * @return JsonModel
     * @throws \P4\Exception
     */
    private function checkStrict()
    {
        $changeId = $this->getEvent()->getRouteMatch()->getParam('change');
        $manager  = new Manager($this->getServiceLocator());
        return $this->translateWorkflowResult($manager->checkStrict($changeId));
    }

    /**
     * Translates the result from a workflow action to the format to be returned by the API
     * @param $result
     * @return JsonModel
     */
    private function translateWorkflowResult($result)
    {
        $logger = $this->getServiceLocator()->get('logger');
        $valid  = $result[Manager::KEY_STATUS] === Manager::STATUS_OK ? true : false;
        if (!$valid) {
            if ($result[Manager::KEY_STATUS] === Manager::STATUS_BAD_CHANGE) {
                // The traditional bad data response
                $this->getResponse()->setStatusCode(Response::STATUS_CODE_404);
            }
        }
        $result = array(
            'isValid'  => $valid,
            'status'   => $result[Manager::KEY_STATUS],
            'messages' => $result[Manager::KEY_MESSAGES]
        );
        $logger->trace('Workflow result ' . var_export($result, true));
        return new JsonModel($result);
    }
}
